use crate::ffi::{
    ffi_arc_dyn_getter, ffi_arc_stringkey_getter, BorrowedPtr, ExternPointer, IdentifiablePointer, NativeResult,
    OwnedPtr,
};
use crate::static_data::{Form, Gender, Species, SpeciesImpl};
use crate::{PkmnError, Random, StringKey};
use hashbrown::HashSet;
use std::ffi::{c_char, CStr};
use std::ptr::drop_in_place;
use std::sync::Arc;

/// Creates a new species.
#[no_mangle]
unsafe extern "C" fn species_new(
    id: u16,
    name: BorrowedPtr<c_char>,
    gender_rate: f32,
    growth_rate: BorrowedPtr<c_char>,
    capture_rate: u8,
    default_form: OwnedPtr<Arc<dyn Form>>,
    flags: *const *const c_char,
    flags_length: usize,
) -> NativeResult<IdentifiablePointer<Arc<dyn Species>>> {
    let name: StringKey = match CStr::from_ptr(name).to_str() {
        Ok(name) => name.into(),
        Err(_) => return NativeResult::err(PkmnError::InvalidCString.into()),
    };
    let growth_rate: StringKey = match CStr::from_ptr(growth_rate).to_str() {
        Ok(growth_rate) => growth_rate.into(),
        Err(_) => return NativeResult::err(PkmnError::InvalidCString.into()),
    };

    let flags = std::slice::from_raw_parts(flags, flags_length);
    let mut flags_set: HashSet<StringKey> = HashSet::with_capacity(flags_length);
    for flag in flags {
        let flag = match CStr::from_ptr(*flag).to_str() {
            Ok(flag) => flag,
            Err(_) => return NativeResult::err(PkmnError::InvalidCString.into()),
        };
        flags_set.insert(flag.into());
    }
    let default_form = match default_form.as_ref() {
        Some(default_form) => default_form,
        None => return NativeResult::err(PkmnError::NullReference.into()),
    };

    let a: Arc<dyn Species> = Arc::new(SpeciesImpl::new(
        id,
        &name,
        gender_rate,
        &growth_rate,
        capture_rate,
        default_form.clone(),
        flags_set,
    ));
    NativeResult::ok(a.into())
}

/// Drop a reference to the species.
#[no_mangle]
unsafe extern "C" fn species_drop(ptr: OwnedPtr<Arc<dyn Species>>) {
    drop_in_place(ptr);
}

ffi_arc_dyn_getter!(Species, id, u16);
ffi_arc_stringkey_getter!(Species, name);
ffi_arc_dyn_getter!(Species, gender_rate, f32);
ffi_arc_stringkey_getter!(Species, growth_rate);
ffi_arc_dyn_getter!(Species, capture_rate, u8);

/// Adds a new form to the species.
#[no_mangle]
unsafe extern "C" fn species_add_form(
    mut species: ExternPointer<Arc<dyn Species>>,
    name: BorrowedPtr<c_char>,
    form: OwnedPtr<Arc<dyn Form>>,
) -> NativeResult<()> {
    let form = match form.as_ref() {
        Some(form) => form.clone(),
        None => return NativeResult::err(PkmnError::NullReference.into()),
    };
    species.as_mut().add_form(CStr::from_ptr(name).into(), form);
    NativeResult::ok(())
}

/// Gets a form by name.
#[no_mangle]
unsafe extern "C" fn species_get_form(
    species: ExternPointer<Arc<dyn Species>>,
    name: BorrowedPtr<c_char>,
) -> IdentifiablePointer<Arc<dyn Form>> {
    let form = species.as_ref().get_form(&CStr::from_ptr(name).into());
    if let Some(form) = form {
        form.into()
    } else {
        IdentifiablePointer::none()
    }
}

/// Gets a form by name.
#[no_mangle]
unsafe extern "C" fn species_get_random_gender(species: ExternPointer<Arc<dyn Species>>, seed: u64) -> Gender {
    let mut rand = Random::new(seed as u128);
    species.as_ref().get_random_gender(&mut rand)
}
